from . import src
from ._version import __version__
from .src import zms_announce as announcer
from .src import zms_types as types
from .src.zms_consts import API_URL
from .src.zms_profiles import get_profile_infos, get_profiles



import re as regex
from collections.abc import Callable, Iterable, Mapping
from functools import reduce

import requests
import tcrutils as tcr

from .zms_consts import API_URL
from .zms_rate import GLOBAL_DEFAULT_RATE_LIMITER, RateLimiter
from .zms_types import Profile, ProfileInfo, ProfileInfoError


def _get_single_profiles(id: int | str, rl: RateLimiter, announcer: Callable) -> list[Profile]:
  url = API_URL / 'profiles' / id

  rl.sleep_until_over()

  response = requests.get(url).json()

  if announcer is not None:
    for profile in response:
      announcer(id, profile)

  return response


def _get_single_profile(id: int | str, rl: RateLimiter, announcer: Callable) -> ProfileInfo | ProfileInfoError:
  url = API_URL / 'profile' / id

  rl.sleep_until_over()

  response = requests.get(url).json()

  if announcer is not None:
    announcer(id, response)

  return response


def get_profiles(
  *ids: int | str,
  rate_limiter=GLOBAL_DEFAULT_RATE_LIMITER,
  announcer: Callable[[Profile], None] | None = None,
) -> list[Profile]:
  if not ids:
    return []

  if not all(tcr.discord.is_snowflake(x, allow_string=True) for x in ids):
    raise ValueError('All ids must be discord Snowflakes. (int-like strings allowed)')

  ids = [int(x) for x in ids]
  ids = list(set(ids))  # Remove duplicates

  requested = [_get_single_profiles(x, rl=rate_limiter, announcer=announcer) for x in ids]

  return reduce(lambda x, y: x + y, requested)


def get_profile_infos(
  *profiles: Profile,
  rate_limiter=GLOBAL_DEFAULT_RATE_LIMITER,
  announcer: Callable[[ProfileInfo | ProfileInfoError], None] | None = None,
  drop_errored_profiles: bool = False,
) -> list[ProfileInfo | ProfileInfoError]:
  if not profiles:
    return []

  if not all(isinstance(x, dict) for x in profiles) or not all('id' in x for x in profiles):
    raise ValueError("All must be valid profile dictionaries (or at least dictionaries containing the key 'id')")

  r = [profile['id'] for profile in profiles]

  r = [_get_single_profile(x, rl=rate_limiter, announcer=announcer) for x in r]

  if drop_errored_profiles:
    r = [x for x in r if 'error' not in x]

  return r



from colored import Back, Fore, Style
from tcrutils import custom_zfill, cut_at

from .zms_types import Profile, ProfileInfo, ProfileInfoError

C_GOLD = Fore.YELLOW + Style.BOLD
C_RED = Fore.RED + Style.BOLD
C_WHITE = Fore.WHITE + Style.BOLD
C_RESET = Style.RESET


def _print_error(id, x):
  print(f"{C_RED}Failed to fetch profile info of {C_WHITE}{id}{C_RED}: {C_WHITE}{x['name']}{C_RED}, {C_WHITE}{x['msg']}{C_RESET}")


def profile_info(id: int, x: ProfileInfo | ProfileInfoError) -> None:
  if 'error' in x:
    _print_error(id, x)
    return

  print(f"{C_GOLD}Fetched profile info {C_WHITE}{f'{x['id']:^30}'}{C_GOLD} - [{C_WHITE}{cut_at(x['name'], 25)}{C_GOLD}]{C_RESET}")


def profiles(id: int, x: Profile) -> None:
  if 'error' in x:
    _print_error(id, x)
    return

  print(f"{C_GOLD}Fetched profile {C_WHITE}{f'{x['id']}'}{C_GOLD}")



import time

ONE_TENTH_OF_CURRENT_RATE_LIMIT_ON_THIS_RATE_LIMITER = object()


class RateLimiter:
  """Rate limiter. Sleep for the duration of or check if the rate limit is over."""

  rate_limit_seconds: int
  """Rate limit in seconds. You may change this value on an initialised rate limiter to change its rate limit."""

  def __init__(self, rate_limit_seconds: int):
    self.rate_limit_seconds = rate_limit_seconds
    self._last_request_unix = 0

  def rate_limited(self):
    """Return False if ready for another request. True otherwise.

    If you get True ("keep waiting") you should periodically call this function again to see if the time for your limit passed.
    You may do other work while waiting.

    Example:
    ```py
    do_request()
    while rate_limiter.rate_limitted():
      do_work()
    do_request()
    ```
    """
    if time.time() - self._last_request_unix < self.rate_limit_seconds:
      return True
    self._last_request_unix = time.time()
    return False

  def sleep_until_over(self, check_every_seconds: int = ONE_TENTH_OF_CURRENT_RATE_LIMIT_ON_THIS_RATE_LIMITER):
    """Sleep until rate limit is passed. You may not do other work while this is ongoing, if you wish to do so use :meth:`rate_limited`."""
    if check_every_seconds is ONE_TENTH_OF_CURRENT_RATE_LIMIT_ON_THIS_RATE_LIMITER:
      check_every_seconds = self.rate_limit_seconds / 10

    while self.rate_limited():
      time.sleep(check_every_seconds)


GLOBAL_DEFAULT_RATE_LIMITER = RateLimiter(1)
